/* $Id$

This file is part of gldbg
Copyright © 2009  Mikko Rasa, Mikkosoft Productions
Distributed under the GPL
*/

#include <stdlib.h>
#include <string.h>
#include "functions.h"
#include "gldecoder.h"
#include "tmpalloc.h"

static unsigned read_short(short *, const char *);
static unsigned read_int(int *, const char *);
static int decode_func(GlDecoder *, unsigned short, const char *);
static int decode_gldfunc(GlDecoder *, unsigned short, const char *);

GlDecoder *gldecoder_new(void *user_data, void (*destroy)(void *))
{
	GlDecoder *dec;

	dec = (GlDecoder *)malloc(sizeof(GlDecoder));
	memset(dec, 0, sizeof(GlDecoder));
	dec->user_data = user_data;
	dec->destroy = destroy;

	return dec;
}

void gldecoder_delete(GlDecoder *dec)
{
	if(dec->destroy)
		dec->destroy(dec->user_data);
	free(dec);
}

int gldecoder_decode(GlDecoder *dec, const char *data, unsigned len)
{
	unsigned pos = 0;
	unsigned pktlen;
	unsigned short func;
	int ret;

	if(len<sizeof(int)+sizeof(short))
		return -1;
	pos += read_int((int *)&pktlen, data);
	if(len<pktlen)
		return -1;
	pos += read_short((short *)&func, data+pos);
	if(dec)
	{
		if(func&0x8000)
			ret = decode_gldfunc(dec, func, data+pos);
		else
			ret = decode_func(dec, func, data+pos);
		if(ret<0)
			return -1;
	}
	return pktlen;
}

static unsigned read_char(char *v, const char *data)
{
	*v = *data;
	return 1;
}

static unsigned read_short(short *v, const char *data)
{
	*v = *(short *)data;
	return sizeof(short);
}

static unsigned read_int(int *v, const char *data)
{
	*v = *(int *)data;
	return sizeof(int);
}

static unsigned read_long(long *v, const char *data)
{
	*v = *(long *)data;
	return sizeof(long);
}

static unsigned read_long_long(long long *v, const char *data)
{
	*v = *(long long *)data;
	return sizeof(long long);
}

static unsigned read_float(float *v, const char *data)
{
	*v = *(float *)data;
	return sizeof(float);
}

static unsigned read_double(double *v, const char *data)
{
	*v = *(double *)data;
	return sizeof(double);
}

typedef void *pointer;

static unsigned read_pointer(pointer *v, const char *data)
{
	*v = *(void **)data;
	return sizeof(void *);
}

static unsigned read_data(const void **v, const char *data)
{
	int vlen;
	unsigned pos = 0;
	pos += read_int(&vlen, data);
	if(vlen)
		*v = data+pos;
	else
		*v = NULL;
	return pos+vlen;
}

typedef const char *string;

static unsigned read_string(string *v, const char *data)
{
	return read_data((const void **)v, data);
}

static unsigned read_string_array(string **v, const char *data)
{
	int count;
	unsigned pos = 0;
	int i;
	pos += read_int(&count, data);
	*v = (string *)tmpalloc(count*sizeof(string));
	for(i=0; i<count; ++i)
		pos += read_string(*v+i, data+pos);
	return pos;
}

#include "gldecoder.funcs"

static int decode_gldError(GlDecoder *dec, const char *data)
{
	unsigned pos = 0;
	GLenum code;
	pos += read_int((int *)&code, data);
	if(dec->gldError)
		dec->gldError(dec->user_data, code);
	return pos;
}

static int decode_gldfunc(GlDecoder *dec, unsigned short func, const char *data)
{
	switch(func)
	{
	case FUNC_GLDERROR: return decode_gldError(dec, data);
	default: return -1;
	}
}
